from __future__ import annotations

from collections.abc import AsyncGenerator, Generator
from concurrent.futures import Future
from datetime import datetime, timezone
from queue import Empty, Queue
from typing import Any, cast

import pytest
from _pytest.fixtures import SubRequest
from _pytest.logging import LogCaptureFixture
from anyio import CancelScope, create_memory_object_stream, fail_after
from pytest_lazyfixture import lazy_fixture

from apscheduler import Event, ScheduleAdded
from apscheduler.abc import AsyncEventBroker, EventBroker, Serializer


@pytest.fixture
def local_broker() -> EventBroker:
    from apscheduler.eventbrokers.local import LocalEventBroker

    return LocalEventBroker()


@pytest.fixture
def local_async_broker() -> AsyncEventBroker:
    from apscheduler.eventbrokers.async_local import LocalAsyncEventBroker

    return LocalAsyncEventBroker()


@pytest.fixture
def redis_broker(serializer: Serializer) -> EventBroker:
    from apscheduler.eventbrokers.redis import RedisEventBroker

    broker = RedisEventBroker.from_url("redis://localhost:6379")
    broker.serializer = serializer
    return broker


@pytest.fixture
def mqtt_broker(serializer: Serializer) -> EventBroker:
    from paho.mqtt.client import Client

    from apscheduler.eventbrokers.mqtt import MQTTEventBroker

    return MQTTEventBroker(Client(), serializer=serializer)


@pytest.fixture
async def asyncpg_broker(serializer: Serializer) -> AsyncEventBroker:
    from asyncpg import create_pool

    from apscheduler.eventbrokers.asyncpg import AsyncpgEventBroker

    pool = await create_pool("postgres://postgres:secret@localhost:5432/testdb")
    broker = AsyncpgEventBroker.from_asyncpg_pool(pool, serializer=serializer)
    yield broker
    await pool.close()


@pytest.fixture(
    params=[
        pytest.param(lazy_fixture("local_broker"), id="local"),
        pytest.param(
            lazy_fixture("redis_broker"),
            id="redis",
            marks=[pytest.mark.external_service],
        ),
        pytest.param(
            lazy_fixture("mqtt_broker"), id="mqtt", marks=[pytest.mark.external_service]
        ),
    ]
)
def broker(request: SubRequest) -> Generator[EventBroker, Any, None]:
    request.param.start()
    yield request.param
    request.param.stop()


@pytest.fixture(
    params=[
        pytest.param(lazy_fixture("local_async_broker"), id="local"),
        pytest.param(
            lazy_fixture("asyncpg_broker"),
            id="asyncpg",
            marks=[pytest.mark.external_service],
        ),
    ]
)
async def raw_async_broker(request: SubRequest) -> AsyncEventBroker:
    return cast(AsyncEventBroker, request.param)


@pytest.fixture
async def async_broker(
    raw_async_broker: AsyncEventBroker,
) -> AsyncGenerator[AsyncEventBroker, Any]:
    await raw_async_broker.start()
    yield raw_async_broker
    await raw_async_broker.stop()


class TestEventBroker:
    def test_publish_subscribe(self, broker: EventBroker) -> None:
        queue: Queue[Event] = Queue()
        broker.subscribe(queue.put_nowait)
        broker.subscribe(queue.put_nowait)
        event = ScheduleAdded(
            schedule_id="schedule1",
            next_fire_time=datetime(2021, 9, 11, 12, 31, 56, 254867, timezone.utc),
        )
        broker.publish(event)
        event1 = queue.get(timeout=3)
        event2 = queue.get(timeout=1)

        assert event1 == event2
        assert isinstance(event1, ScheduleAdded)
        assert isinstance(event1.timestamp, datetime)
        assert event1.schedule_id == "schedule1"
        assert event1.next_fire_time == datetime(
            2021, 9, 11, 12, 31, 56, 254867, timezone.utc
        )

    def test_subscribe_one_shot(self, broker: EventBroker) -> None:
        queue: Queue[Event] = Queue()
        broker.subscribe(queue.put_nowait, one_shot=True)
        event = ScheduleAdded(
            schedule_id="schedule1",
            next_fire_time=datetime(2021, 9, 11, 12, 31, 56, 254867, timezone.utc),
        )
        broker.publish(event)
        event = ScheduleAdded(
            schedule_id="schedule2",
            next_fire_time=datetime(2021, 9, 12, 8, 42, 11, 968481, timezone.utc),
        )
        broker.publish(event)
        received_event = queue.get(timeout=3)
        with pytest.raises(Empty):
            queue.get(timeout=0.1)

        assert isinstance(received_event, ScheduleAdded)
        assert received_event.schedule_id == "schedule1"

    def test_unsubscribe(self, broker: EventBroker, caplog) -> None:
        queue: Queue[Event] = Queue()
        subscription = broker.subscribe(queue.put_nowait)
        broker.publish(Event())
        queue.get(timeout=3)

        subscription.unsubscribe()
        broker.publish(Event())
        with pytest.raises(Empty):
            queue.get(timeout=0.1)

    def test_publish_no_subscribers(
        self, broker: EventBroker, caplog: LogCaptureFixture
    ) -> None:
        broker.publish(Event())
        assert not caplog.text

    def test_publish_exception(
        self, broker: EventBroker, caplog: LogCaptureFixture
    ) -> None:
        def bad_subscriber(event: Event) -> None:
            raise Exception("foo")

        timestamp = datetime.now(timezone.utc)
        event_future: Future[Event] = Future()
        broker.subscribe(bad_subscriber)
        broker.subscribe(event_future.set_result)
        broker.publish(Event(timestamp=timestamp))

        event = event_future.result(3)
        assert isinstance(event, Event)
        assert event.timestamp == timestamp
        assert "Error delivering Event" in caplog.text


@pytest.mark.anyio
class TestAsyncEventBroker:
    async def test_publish_subscribe(self, async_broker: AsyncEventBroker) -> None:
        send, receive = create_memory_object_stream(2)
        async_broker.subscribe(send.send)
        async_broker.subscribe(send.send_nowait)
        event = ScheduleAdded(
            schedule_id="schedule1",
            next_fire_time=datetime(2021, 9, 11, 12, 31, 56, 254867, timezone.utc),
        )
        await async_broker.publish(event)

        with fail_after(3):
            event1 = await receive.receive()
            event2 = await receive.receive()

        assert event1 == event2
        assert isinstance(event1, ScheduleAdded)
        assert isinstance(event1.timestamp, datetime)
        assert event1.schedule_id == "schedule1"
        assert event1.next_fire_time == datetime(
            2021, 9, 11, 12, 31, 56, 254867, timezone.utc
        )

    async def test_subscribe_one_shot(self, async_broker: AsyncEventBroker) -> None:
        send, receive = create_memory_object_stream(2)
        async_broker.subscribe(send.send, one_shot=True)
        event = ScheduleAdded(
            schedule_id="schedule1",
            next_fire_time=datetime(2021, 9, 11, 12, 31, 56, 254867, timezone.utc),
        )
        await async_broker.publish(event)
        event = ScheduleAdded(
            schedule_id="schedule2",
            next_fire_time=datetime(2021, 9, 12, 8, 42, 11, 968481, timezone.utc),
        )
        await async_broker.publish(event)

        with fail_after(3):
            received_event = await receive.receive()

        with pytest.raises(TimeoutError), fail_after(0.1):
            await receive.receive()

        assert isinstance(received_event, ScheduleAdded)
        assert received_event.schedule_id == "schedule1"

    async def test_unsubscribe(self, async_broker: AsyncEventBroker) -> None:
        send, receive = create_memory_object_stream()
        subscription = async_broker.subscribe(send.send)
        await async_broker.publish(Event())
        with fail_after(3):
            await receive.receive()

        subscription.unsubscribe()
        await async_broker.publish(Event())
        with pytest.raises(TimeoutError), fail_after(0.1):
            await receive.receive()

    async def test_publish_no_subscribers(
        self, async_broker: AsyncEventBroker, caplog: LogCaptureFixture
    ) -> None:
        await async_broker.publish(Event())
        assert not caplog.text

    async def test_publish_exception(
        self, async_broker: AsyncEventBroker, caplog: LogCaptureFixture
    ) -> None:
        def bad_subscriber(event: Event) -> None:
            raise Exception("foo")

        timestamp = datetime.now(timezone.utc)
        send, receive = create_memory_object_stream()
        async_broker.subscribe(bad_subscriber)
        async_broker.subscribe(send.send)
        await async_broker.publish(Event(timestamp=timestamp))

        received_event = await receive.receive()
        assert received_event.timestamp == timestamp
        assert "Error delivering Event" in caplog.text

    async def test_cancel_start(self, raw_async_broker: AsyncEventBroker) -> None:
        with CancelScope() as scope:
            scope.cancel()
            await raw_async_broker.start()
            await raw_async_broker.stop()

    async def test_cancel_stop(self, raw_async_broker: AsyncEventBroker) -> None:
        with CancelScope() as scope:
            await raw_async_broker.start()
            scope.cancel()
            await raw_async_broker.stop()



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, AsyncContextManager, AsyncGenerator, Callable

import attrs
from anyio import TASK_STATUS_IGNORED, CancelScope, sleep
from asyncpg import Connection
from asyncpg.pool import Pool

from .._events import Event
from .._exceptions import SerializationError
from ..abc import Serializer
from ..serializers.json import JSONSerializer
from .async_local import LocalAsyncEventBroker
from .base import DistributedEventBrokerMixin

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncEngine


@attrs.define(eq=False)
class AsyncpgEventBroker(LocalAsyncEventBroker, DistributedEventBrokerMixin):
    """
    An asynchronous, asyncpg_ based event broker that uses a PostgreSQL server to
    broadcast events using its ``NOTIFY`` mechanism.

    .. _asyncpg: https://pypi.org/project/asyncpg/

    :param connection_factory: a callable that creates an async context manager that
        yields a new asyncpg connection
    :param serializer: the serializer used to (de)serialize events for transport
    :param channel: the ``NOTIFY`` channel to use
    :param max_idle_time: maximum time to let the connection go idle, before sending a
        ``SELECT 1`` query to prevent a connection timeout
    """

    connection_factory: Callable[[], AsyncContextManager[Connection]]
    serializer: Serializer = attrs.field(kw_only=True, factory=JSONSerializer)
    channel: str = attrs.field(kw_only=True, default="apscheduler")
    max_idle_time: float = attrs.field(kw_only=True, default=30)
    _listen_cancel_scope: CancelScope = attrs.field(init=False)

    @classmethod
    def from_asyncpg_pool(cls, pool: Pool, **kwargs) -> AsyncpgEventBroker:
        """
        Create a new asyncpg event broker from an existing asyncpg connection pool.

        :param pool: an asyncpg connection pool
        :param kwargs: keyword arguments to pass to the initializer of this class
        :return: the newly created event broker

        """
        return cls(pool.acquire, **kwargs)

    @classmethod
    def from_async_sqla_engine(
        cls, engine: AsyncEngine, **kwargs
    ) -> AsyncpgEventBroker:
        """
        Create a new asyncpg event broker from an SQLAlchemy engine.

        :param engine: an asynchronous SQLAlchemy engine using asyncpg as the driver
        :type engine: ~sqlalchemy.ext.asyncio.AsyncEngine
        :param kwargs: keyword arguments to pass to the initializer of this class
        :return: the newly created event broker

        """
        if engine.dialect.driver != "asyncpg":
            raise ValueError(
                f'The driver in the engine must be "asyncpg" (current: '
                f"{engine.dialect.driver})"
            )

        @asynccontextmanager
        async def connection_factory() -> AsyncGenerator[Connection, None]:
            conn = await engine.raw_connection()
            try:
                yield conn.connection._connection
            finally:
                conn.close()

        return cls(connection_factory, **kwargs)

    async def start(self) -> None:
        await super().start()
        self._listen_cancel_scope = await self._task_group.start(
            self._listen_notifications
        )

    async def stop(self, *, force: bool = False) -> None:
        self._listen_cancel_scope.cancel()
        await super().stop(force=force)

    async def _listen_notifications(self, *, task_status=TASK_STATUS_IGNORED) -> None:
        def callback(connection, pid, channel: str, payload: str) -> None:
            event = self.reconstitute_event_str(payload)
            if event is not None:
                self._task_group.start_soon(self.publish_local, event)

        task_started_sent = False
        with CancelScope() as cancel_scope:
            while True:
                async with self.connection_factory() as conn:
                    await conn.add_listener(self.channel, callback)
                    if not task_started_sent:
                        task_status.started(cancel_scope)
                        task_started_sent = True

                    try:
                        while True:
                            await sleep(self.max_idle_time)
                            await conn.execute("SELECT 1")
                    finally:
                        await conn.remove_listener(self.channel, callback)

    async def publish(self, event: Event) -> None:
        notification = self.generate_notification_str(event)
        if len(notification) > 7999:
            raise SerializationError(
                "Serialized event object exceeds 7999 bytes in size"
            )

        async with self.connection_factory() as conn:
            await conn.execute("SELECT pg_notify($1, $2)", self.channel, notification)
            return



"""
Example demonstrating the separation of scheduler and worker.
This script runs the scheduler part. You need to be running both this and the worker
script simultaneously in order for the scheduled task to be run.

Requires the "postgresql" and "redis" services to be running.
To install prerequisites: pip install sqlalchemy psycopg2 redis
To run: python sync_scheduler.py

When run together with sync_worker.py, it should print a line on the console
on a one-second interval.
"""

from __future__ import annotations

from example_tasks import tick
from sqlalchemy.future import create_engine

from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore
from apscheduler.eventbrokers.redis import RedisEventBroker
from apscheduler.schedulers.sync import Scheduler
from apscheduler.triggers.interval import IntervalTrigger

engine = create_engine("postgresql+psycopg2://postgres:secret@localhost/testdb")
data_store = SQLAlchemyDataStore(engine)
event_broker = RedisEventBroker.from_url("redis://localhost")
with Scheduler(data_store, event_broker, start_worker=False) as scheduler:
    scheduler.add_schedule(tick, IntervalTrigger(seconds=1), id="tick")
    scheduler.wait_until_stopped()



from __future__ import annotations

from datetime import datetime

from apscheduler.triggers.date import DateTrigger


def test_run_time(timezone, serializer):
    run_time = datetime(2020, 5, 14, 11, 56, 12, tzinfo=timezone)
    trigger = DateTrigger(run_time)
    if serializer:
        payload = serializer.serialize(trigger)
        trigger = serializer.deserialize(payload)

    assert trigger.next() == run_time
    assert trigger.next() is None
    assert repr(trigger) == "DateTrigger('2020-05-14 11:56:12+02:00')"



from __future__ import annotations

from asyncio import iscoroutinefunction
from concurrent.futures import ThreadPoolExecutor
from contextlib import ExitStack
from threading import Lock
from typing import Any, Callable, Iterable

import attrs

from .._events import Event
from ..abc import EventBroker, Subscription
from .base import BaseEventBroker


@attrs.define(eq=False)
class LocalEventBroker(EventBroker, BaseEventBroker):
    """
    Synchronous, local event broker.

    This event broker only broadcasts within the process it runs in, and is therefore
    not suitable for multi-node or multiprocess use cases.

    Does not serialize events.
    """

    _executor: ThreadPoolExecutor = attrs.field(init=False)
    _exit_stack: ExitStack = attrs.field(init=False)
    _subscriptions_lock: Lock = attrs.field(init=False, factory=Lock)

    def start(self) -> None:
        self._executor = ThreadPoolExecutor(1)

    def stop(self, *, force: bool = False) -> None:
        self._executor.shutdown(wait=not force)
        del self._executor

    def subscribe(
        self,
        callback: Callable[[Event], Any],
        event_types: Iterable[type[Event]] | None = None,
        *,
        one_shot: bool = False,
    ) -> Subscription:
        if iscoroutinefunction(callback):
            raise ValueError(
                "Coroutine functions are not supported as callbacks on a synchronous "
                "event source"
            )

        with self._subscriptions_lock:
            return super().subscribe(callback, event_types, one_shot=one_shot)

    def unsubscribe(self, token: object) -> None:
        with self._subscriptions_lock:
            super().unsubscribe(token)

    def publish(self, event: Event) -> None:
        self.publish_local(event)

    def publish_local(self, event: Event) -> None:
        event_type = type(event)
        with self._subscriptions_lock:
            one_shot_tokens: list[object] = []
            for _token, subscription in self._subscriptions.items():
                if (
                    subscription.event_types is None
                    or event_type in subscription.event_types
                ):
                    self._executor.submit(
                        self._deliver_event, subscription.callback, event
                    )
                    if subscription.one_shot:
                        one_shot_tokens.append(subscription.token)

            for token in one_shot_tokens:
                super().unsubscribe(token)

    def _deliver_event(self, func: Callable[[Event], Any], event: Event) -> None:
        try:
            func(event)
        except BaseException:
            self._logger.exception(
                "Error delivering %s event", event.__class__.__name__
            )



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any

import attrs

from .._validators import as_aware_datetime, require_state_version
from ..abc import Trigger
from ..marshalling import marshal_date, unmarshal_date


@attrs.define(kw_only=True)
class IntervalTrigger(Trigger):
    """
    Triggers on specified intervals.

    The first trigger time is on ``start_time`` which is the  moment the trigger was
    created unless specifically overridden. If ``end_time`` is specified, the last
    trigger time will be at or before that time. If no ``end_time`` has been given, the
    trigger will produce new trigger times as long as the resulting datetimes are valid
    datetimes in Python.

    :param weeks: number of weeks to wait
    :param days: number of days to wait
    :param hours: number of hours to wait
    :param minutes: number of minutes to wait
    :param seconds: number of seconds to wait
    :param microseconds: number of microseconds to wait
    :param start_time: first trigger date/time (defaults to current date/time if
        omitted)
    :param end_time: latest possible date/time to trigger on
    """

    weeks: float = 0
    days: float = 0
    hours: float = 0
    minutes: float = 0
    seconds: float = 0
    microseconds: float = 0
    start_time: datetime = attrs.field(
        converter=as_aware_datetime, factory=datetime.now
    )
    end_time: datetime | None = attrs.field(converter=as_aware_datetime, default=None)
    _interval: timedelta = attrs.field(init=False, eq=False, repr=False)
    _last_fire_time: datetime | None = attrs.field(init=False, eq=False, default=None)

    def __attrs_post_init__(self) -> None:
        self._interval = timedelta(
            weeks=self.weeks,
            days=self.days,
            hours=self.hours,
            minutes=self.minutes,
            seconds=self.seconds,
            microseconds=self.microseconds,
        )

        if self._interval.total_seconds() <= 0:
            raise ValueError("The time interval must be positive")

        if self.end_time and self.end_time < self.start_time:
            raise ValueError("end_time cannot be earlier than start_time")

    def next(self) -> datetime | None:
        if self._last_fire_time is None:
            self._last_fire_time = self.start_time
        else:
            self._last_fire_time += self._interval

        if self.end_time is None or self._last_fire_time <= self.end_time:
            return self._last_fire_time
        else:
            return None

    def __getstate__(self) -> dict[str, Any]:
        return {
            "version": 1,
            "interval": [
                self.weeks,
                self.days,
                self.hours,
                self.minutes,
                self.seconds,
                self.microseconds,
            ],
            "start_time": marshal_date(self.start_time),
            "end_time": marshal_date(self.end_time),
            "last_fire_time": marshal_date(self._last_fire_time),
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        require_state_version(self, state, 1)
        (
            self.weeks,
            self.days,
            self.hours,
            self.minutes,
            self.seconds,
            self.microseconds,
        ) = state["interval"]
        self.start_time = unmarshal_date(state["start_time"])
        self.end_time = unmarshal_date(state["end_time"])
        self._last_fire_time = unmarshal_date(state["last_fire_time"])
        self._interval = timedelta(
            weeks=self.weeks,
            days=self.days,
            hours=self.hours,
            minutes=self.minutes,
            seconds=self.seconds,
            microseconds=self.microseconds,
        )

    def __repr__(self) -> str:
        fields = []
        for field in "weeks", "days", "hours", "minutes", "seconds", "microseconds":
            value = getattr(self, field)
            if value > 0:
                fields.append(f"{field}={value}")

        fields.append(f"start_time='{self.start_time}'")
        if self.end_time:
            fields.append(f"end_time='{self.end_time}'")

        return f'{self.__class__.__name__}({", ".join(fields)})'



from __future__ import annotations

from collections import defaultdict
from datetime import datetime, timedelta, timezone
from typing import Any, Iterable
from uuid import UUID

import anyio
import attrs
import sniffio
import tenacity
from sqlalchemy import and_, bindparam, or_, select
from sqlalchemy.engine import URL, Result
from sqlalchemy.exc import IntegrityError, InterfaceError
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from sqlalchemy.sql.ddl import DropTable
from sqlalchemy.sql.elements import BindParameter

from .._enums import ConflictPolicy
from .._events import (
    DataStoreEvent,
    JobAcquired,
    JobAdded,
    JobDeserializationFailed,
    ScheduleAdded,
    ScheduleDeserializationFailed,
    ScheduleRemoved,
    ScheduleUpdated,
    TaskAdded,
    TaskRemoved,
    TaskUpdated,
)
from .._exceptions import ConflictingIdError, SerializationError, TaskLookupError
from .._structures import Job, JobResult, Schedule, Task
from ..abc import AsyncEventBroker
from ..marshalling import callable_to_ref
from .base import BaseAsyncDataStore
from .sqlalchemy import _BaseSQLAlchemyDataStore


@attrs.define(eq=False)
class AsyncSQLAlchemyDataStore(_BaseSQLAlchemyDataStore, BaseAsyncDataStore):
    """
    Uses a relational database to store data.

    When started, this data store creates the appropriate tables on the given database
    if they're not already present.

    Operations are retried (in accordance to ``retry_settings``) when an operation
    raises :exc:`sqlalchemy.OperationalError`.

    This store has been tested to work with PostgreSQL (asyncpg driver) and MySQL
    (asyncmy driver).

    :param engine: an asynchronous SQLAlchemy engine
    :param schema: a database schema name to use, if not the default
    :param serializer: the serializer used to (de)serialize tasks, schedules and jobs
        for storage
    :param lock_expiration_delay: maximum amount of time (in seconds) that a scheduler
        or worker can keep a lock on a schedule or task
    :param retry_settings: Tenacity settings for retrying operations in case of a
        database connecitivty problem
    :param start_from_scratch: erase all existing data during startup (useful for test
        suites)
    """

    engine: AsyncEngine

    @classmethod
    def from_url(cls, url: str | URL, **options) -> AsyncSQLAlchemyDataStore:
        """
        Create a new asynchronous SQLAlchemy data store.

        :param url: an SQLAlchemy URL to pass to :func:`~sqlalchemy.create_engine`
            (must use an async dialect like ``asyncpg`` or ``asyncmy``)
        :param kwargs: keyword arguments to pass to the initializer of this class
        :return: the newly created data store

        """
        engine = create_async_engine(url, future=True)
        return cls(engine, **options)

    def _retry(self) -> tenacity.AsyncRetrying:
        # OSError is raised by asyncpg if it can't connect
        return tenacity.AsyncRetrying(
            stop=self.retry_settings.stop,
            wait=self.retry_settings.wait,
            retry=tenacity.retry_if_exception_type((InterfaceError, OSError)),
            after=self._after_attempt,
            sleep=anyio.sleep,
            reraise=True,
        )

    async def start(self, event_broker: AsyncEventBroker) -> None:
        await super().start(event_broker)

        asynclib = sniffio.current_async_library() or "(unknown)"
        if asynclib != "asyncio":
            raise RuntimeError(
                f"This data store requires asyncio; currently running: {asynclib}"
            )

        # Verify that the schema is in place
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    if self.start_from_scratch:
                        for table in self._metadata.sorted_tables:
                            await conn.execute(DropTable(table, if_exists=True))

                    await conn.run_sync(self._metadata.create_all)
                    query = select(self.t_metadata.c.schema_version)
                    result = await conn.execute(query)
                    version = result.scalar()
                    if version is None:
                        await conn.execute(
                            self.t_metadata.insert(values={"schema_version": 1})
                        )
                    elif version > 1:
                        raise RuntimeError(
                            f"Unexpected schema version ({version}); "
                            f"only version 1 is supported by this version of "
                            f"APScheduler"
                        )

    async def _deserialize_schedules(self, result: Result) -> list[Schedule]:
        schedules: list[Schedule] = []
        for row in result:
            try:
                schedules.append(Schedule.unmarshal(self.serializer, row._asdict()))
            except SerializationError as exc:
                await self._events.publish(
                    ScheduleDeserializationFailed(schedule_id=row["id"], exception=exc)
                )

        return schedules

    async def _deserialize_jobs(self, result: Result) -> list[Job]:
        jobs: list[Job] = []
        for row in result:
            try:
                jobs.append(Job.unmarshal(self.serializer, row._asdict()))
            except SerializationError as exc:
                await self._events.publish(
                    JobDeserializationFailed(job_id=row["id"], exception=exc)
                )

        return jobs

    async def add_task(self, task: Task) -> None:
        insert = self.t_tasks.insert().values(
            id=task.id,
            func=callable_to_ref(task.func),
            max_running_jobs=task.max_running_jobs,
            misfire_grace_time=task.misfire_grace_time,
        )
        try:
            async for attempt in self._retry():
                with attempt:
                    async with self.engine.begin() as conn:
                        await conn.execute(insert)
        except IntegrityError:
            update = (
                self.t_tasks.update()
                .values(
                    func=callable_to_ref(task.func),
                    max_running_jobs=task.max_running_jobs,
                    misfire_grace_time=task.misfire_grace_time,
                )
                .where(self.t_tasks.c.id == task.id)
            )
            async for attempt in self._retry():
                with attempt:
                    async with self.engine.begin() as conn:
                        await conn.execute(update)

            await self._events.publish(TaskUpdated(task_id=task.id))
        else:
            await self._events.publish(TaskAdded(task_id=task.id))

    async def remove_task(self, task_id: str) -> None:
        delete = self.t_tasks.delete().where(self.t_tasks.c.id == task_id)
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    result = await conn.execute(delete)
                    if result.rowcount == 0:
                        raise TaskLookupError(task_id)
                    else:
                        await self._events.publish(TaskRemoved(task_id=task_id))

    async def get_task(self, task_id: str) -> Task:
        query = select(
            [
                self.t_tasks.c.id,
                self.t_tasks.c.func,
                self.t_tasks.c.max_running_jobs,
                self.t_tasks.c.state,
                self.t_tasks.c.misfire_grace_time,
            ]
        ).where(self.t_tasks.c.id == task_id)
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    result = await conn.execute(query)
                    row = result.first()

        if row:
            return Task.unmarshal(self.serializer, row._asdict())
        else:
            raise TaskLookupError(task_id)

    async def get_tasks(self) -> list[Task]:
        query = select(
            [
                self.t_tasks.c.id,
                self.t_tasks.c.func,
                self.t_tasks.c.max_running_jobs,
                self.t_tasks.c.state,
                self.t_tasks.c.misfire_grace_time,
            ]
        ).order_by(self.t_tasks.c.id)
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    result = await conn.execute(query)
                    tasks = [
                        Task.unmarshal(self.serializer, row._asdict()) for row in result
                    ]
                    return tasks

    async def add_schedule(
        self, schedule: Schedule, conflict_policy: ConflictPolicy
    ) -> None:
        event: DataStoreEvent
        values = schedule.marshal(self.serializer)
        insert = self.t_schedules.insert().values(**values)
        try:
            async for attempt in self._retry():
                with attempt:
                    async with self.engine.begin() as conn:
                        await conn.execute(insert)
        except IntegrityError:
            if conflict_policy is ConflictPolicy.exception:
                raise ConflictingIdError(schedule.id) from None
            elif conflict_policy is ConflictPolicy.replace:
                del values["id"]
                update = (
                    self.t_schedules.update()
                    .where(self.t_schedules.c.id == schedule.id)
                    .values(**values)
                )
                async for attempt in self._retry():
                    with attempt:
                        async with self.engine.begin() as conn:
                            await conn.execute(update)

                event = ScheduleUpdated(
                    schedule_id=schedule.id, next_fire_time=schedule.next_fire_time
                )
                await self._events.publish(event)
        else:
            event = ScheduleAdded(
                schedule_id=schedule.id, next_fire_time=schedule.next_fire_time
            )
            await self._events.publish(event)

    async def remove_schedules(self, ids: Iterable[str]) -> None:
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    delete = self.t_schedules.delete().where(
                        self.t_schedules.c.id.in_(ids)
                    )
                    if self._supports_update_returning:
                        delete = delete.returning(self.t_schedules.c.id)
                        removed_ids: Iterable[str] = [
                            row[0] for row in await conn.execute(delete)
                        ]
                    else:
                        # TODO: actually check which rows were deleted?
                        await conn.execute(delete)
                        removed_ids = ids

        for schedule_id in removed_ids:
            await self._events.publish(ScheduleRemoved(schedule_id=schedule_id))

    async def get_schedules(self, ids: set[str] | None = None) -> list[Schedule]:
        query = self.t_schedules.select().order_by(self.t_schedules.c.id)
        if ids:
            query = query.where(self.t_schedules.c.id.in_(ids))

        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    result = await conn.execute(query)
                    return await self._deserialize_schedules(result)

    async def acquire_schedules(self, scheduler_id: str, limit: int) -> list[Schedule]:
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    now = datetime.now(timezone.utc)
                    acquired_until = now + timedelta(seconds=self.lock_expiration_delay)
                    schedules_cte = (
                        select(self.t_schedules.c.id)
                        .where(
                            and_(
                                self.t_schedules.c.next_fire_time.isnot(None),
                                self.t_schedules.c.next_fire_time <= now,
                                or_(
                                    self.t_schedules.c.acquired_until.is_(None),
                                    self.t_schedules.c.acquired_until < now,
                                ),
                            )
                        )
                        .order_by(self.t_schedules.c.next_fire_time)
                        .limit(limit)
                        .with_for_update(skip_locked=True)
                        .cte()
                    )
                    subselect = select([schedules_cte.c.id])
                    update = (
                        self.t_schedules.update()
                        .where(self.t_schedules.c.id.in_(subselect))
                        .values(acquired_by=scheduler_id, acquired_until=acquired_until)
                    )
                    if self._supports_update_returning:
                        update = update.returning(*self.t_schedules.columns)
                        result = await conn.execute(update)
                    else:
                        await conn.execute(update)
                        query = self.t_schedules.select().where(
                            and_(self.t_schedules.c.acquired_by == scheduler_id)
                        )
                        result = await conn.execute(query)

                    schedules = await self._deserialize_schedules(result)

        return schedules

    async def release_schedules(
        self, scheduler_id: str, schedules: list[Schedule]
    ) -> None:
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    update_events: list[ScheduleUpdated] = []
                    finished_schedule_ids: list[str] = []
                    update_args: list[dict[str, Any]] = []
                    for schedule in schedules:
                        if schedule.next_fire_time is not None:
                            try:
                                serialized_trigger = self.serializer.serialize(
                                    schedule.trigger
                                )
                            except SerializationError:
                                self._logger.exception(
                                    "Error serializing trigger for schedule %r – "
                                    "removing from data store",
                                    schedule.id,
                                )
                                finished_schedule_ids.append(schedule.id)
                                continue

                            update_args.append(
                                {
                                    "p_id": schedule.id,
                                    "p_trigger": serialized_trigger,
                                    "p_next_fire_time": schedule.next_fire_time,
                                }
                            )
                        else:
                            finished_schedule_ids.append(schedule.id)

                    # Update schedules that have a next fire time
                    if update_args:
                        p_id: BindParameter = bindparam("p_id")
                        p_trigger: BindParameter = bindparam("p_trigger")
                        p_next_fire_time: BindParameter = bindparam("p_next_fire_time")
                        update = (
                            self.t_schedules.update()
                            .where(
                                and_(
                                    self.t_schedules.c.id == p_id,
                                    self.t_schedules.c.acquired_by == scheduler_id,
                                )
                            )
                            .values(
                                trigger=p_trigger,
                                next_fire_time=p_next_fire_time,
                                acquired_by=None,
                                acquired_until=None,
                            )
                        )
                        next_fire_times = {
                            arg["p_id"]: arg["p_next_fire_time"] for arg in update_args
                        }
                        # TODO: actually check which rows were updated?
                        await conn.execute(update, update_args)
                        updated_ids = list(next_fire_times)

                        for schedule_id in updated_ids:
                            event = ScheduleUpdated(
                                schedule_id=schedule_id,
                                next_fire_time=next_fire_times[schedule_id],
                            )
                            update_events.append(event)

                    # Remove schedules that have no next fire time or failed to
                    # serialize
                    if finished_schedule_ids:
                        delete = self.t_schedules.delete().where(
                            self.t_schedules.c.id.in_(finished_schedule_ids)
                        )
                        await conn.execute(delete)

        for event in update_events:
            await self._events.publish(event)

        for schedule_id in finished_schedule_ids:
            await self._events.publish(ScheduleRemoved(schedule_id=schedule_id))

    async def get_next_schedule_run_time(self) -> datetime | None:
        statenent = (
            select(self.t_schedules.c.next_fire_time)
            .where(self.t_schedules.c.next_fire_time.isnot(None))
            .order_by(self.t_schedules.c.next_fire_time)
            .limit(1)
        )
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    result = await conn.execute(statenent)
                    return result.scalar()

    async def add_job(self, job: Job) -> None:
        marshalled = job.marshal(self.serializer)
        insert = self.t_jobs.insert().values(**marshalled)
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    await conn.execute(insert)

        event = JobAdded(
            job_id=job.id,
            task_id=job.task_id,
            schedule_id=job.schedule_id,
            tags=job.tags,
        )
        await self._events.publish(event)

    async def get_jobs(self, ids: Iterable[UUID] | None = None) -> list[Job]:
        query = self.t_jobs.select().order_by(self.t_jobs.c.id)
        if ids:
            job_ids = [job_id for job_id in ids]
            query = query.where(self.t_jobs.c.id.in_(job_ids))

        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    result = await conn.execute(query)
                    return await self._deserialize_jobs(result)

    async def acquire_jobs(self, worker_id: str, limit: int | None = None) -> list[Job]:
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    now = datetime.now(timezone.utc)
                    acquired_until = now + timedelta(seconds=self.lock_expiration_delay)
                    query = (
                        self.t_jobs.select()
                        .join(self.t_tasks, self.t_tasks.c.id == self.t_jobs.c.task_id)
                        .where(
                            or_(
                                self.t_jobs.c.acquired_until.is_(None),
                                self.t_jobs.c.acquired_until < now,
                            )
                        )
                        .order_by(self.t_jobs.c.created_at)
                        .with_for_update(skip_locked=True)
                        .limit(limit)
                    )

                    result = await conn.execute(query)
                    if not result:
                        return []

                    # Mark the jobs as acquired by this worker
                    jobs = await self._deserialize_jobs(result)
                    task_ids: set[str] = {job.task_id for job in jobs}

                    # Retrieve the limits
                    query = select(
                        [
                            self.t_tasks.c.id,
                            self.t_tasks.c.max_running_jobs
                            - self.t_tasks.c.running_jobs,
                        ]
                    ).where(
                        self.t_tasks.c.max_running_jobs.isnot(None),
                        self.t_tasks.c.id.in_(task_ids),
                    )
                    result = await conn.execute(query)
                    job_slots_left: dict[str, int] = dict(result.fetchall())

                    # Filter out jobs that don't have free slots
                    acquired_jobs: list[Job] = []
                    increments: dict[str, int] = defaultdict(lambda: 0)
                    for job in jobs:
                        # Don't acquire the job if there are no free slots left
                        slots_left = job_slots_left.get(job.task_id)
                        if slots_left == 0:
                            continue
                        elif slots_left is not None:
                            job_slots_left[job.task_id] -= 1

                        acquired_jobs.append(job)
                        increments[job.task_id] += 1

                    if acquired_jobs:
                        # Mark the acquired jobs as acquired by this worker
                        acquired_job_ids = [job.id for job in acquired_jobs]
                        update = (
                            self.t_jobs.update()
                            .values(
                                acquired_by=worker_id, acquired_until=acquired_until
                            )
                            .where(self.t_jobs.c.id.in_(acquired_job_ids))
                        )
                        await conn.execute(update)

                        # Increment the running job counters on each task
                        p_id: BindParameter = bindparam("p_id")
                        p_increment: BindParameter = bindparam("p_increment")
                        params = [
                            {"p_id": task_id, "p_increment": increment}
                            for task_id, increment in increments.items()
                        ]
                        update = (
                            self.t_tasks.update()
                            .values(
                                running_jobs=self.t_tasks.c.running_jobs + p_increment
                            )
                            .where(self.t_tasks.c.id == p_id)
                        )
                        await conn.execute(update, params)

        # Publish the appropriate events
        for job in acquired_jobs:
            await self._events.publish(JobAcquired(job_id=job.id, worker_id=worker_id))

        return acquired_jobs

    async def release_job(
        self, worker_id: str, task_id: str, result: JobResult
    ) -> None:
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    # Record the job result
                    if result.expires_at > result.finished_at:
                        marshalled = result.marshal(self.serializer)
                        insert = self.t_job_results.insert().values(**marshalled)
                        await conn.execute(insert)

                    # Decrement the number of running jobs for this task
                    update = (
                        self.t_tasks.update()
                        .values(running_jobs=self.t_tasks.c.running_jobs - 1)
                        .where(self.t_tasks.c.id == task_id)
                    )
                    await conn.execute(update)

                    # Delete the job
                    delete = self.t_jobs.delete().where(
                        self.t_jobs.c.id == result.job_id
                    )
                    await conn.execute(delete)

    async def get_job_result(self, job_id: UUID) -> JobResult | None:
        async for attempt in self._retry():
            with attempt:
                async with self.engine.begin() as conn:
                    # Retrieve the result
                    query = self.t_job_results.select().where(
                        self.t_job_results.c.job_id == job_id
                    )
                    row = (await conn.execute(query)).first()

                    # Delete the result
                    delete = self.t_job_results.delete().where(
                        self.t_job_results.c.job_id == job_id
                    )
                    await conn.execute(delete)

                    return (
                        JobResult.unmarshal(self.serializer, row._asdict())
                        if row
                        else None
                    )



from __future__ import annotations

from apscheduler.abc import (
    AsyncDataStore,
    AsyncEventBroker,
    DataStore,
    EventBroker,
    EventSource,
)


class BaseDataStore(DataStore):
    _events: EventBroker

    def start(self, event_broker: EventBroker) -> None:
        self._events = event_broker

    def stop(self, *, force: bool = False) -> None:
        del self._events

    @property
    def events(self) -> EventSource:
        return self._events


class BaseAsyncDataStore(AsyncDataStore):
    _events: AsyncEventBroker

    async def start(self, event_broker: AsyncEventBroker) -> None:
        self._events = event_broker

    async def stop(self, *, force: bool = False) -> None:
        del self._events

    @property
    def events(self) -> EventSource:
        return self._events



"""
Example demonstrating use with WSGI (raw WSGI application, no framework).

Requires the "postgresql" service to be running.
To install prerequisites: pip install sqlalchemy psycopg2 uwsgi
To run: uwsgi -T --http :8000 --wsgi-file wsgi_noframework.py

It should print a line on the console on a one-second interval while running a
basic web app at http://localhost:8000.
"""

from __future__ import annotations

from datetime import datetime

from sqlalchemy.future import create_engine

from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore
from apscheduler.schedulers.sync import Scheduler
from apscheduler.triggers.interval import IntervalTrigger


def tick():
    print("Hello, the time is", datetime.now())


def application(environ, start_response):
    response_body = b"Hello, World!"
    response_headers = [
        ("Content-Type", "text/plain"),
        ("Content-Length", str(len(response_body))),
    ]
    start_response("200 OK", response_headers)
    return [response_body]


engine = create_engine("postgresql+psycopg2://postgres:secret@localhost/testdb")
data_store = SQLAlchemyDataStore(engine)
scheduler = Scheduler(data_store)
scheduler.add_schedule(tick, IntervalTrigger(seconds=1), id="tick")
scheduler.start_in_background()



from __future__ import annotations

import atexit
import os
import platform
import random
import threading
from concurrent.futures import Future
from contextlib import ExitStack
from datetime import datetime, timedelta, timezone
from logging import Logger, getLogger
from types import TracebackType
from typing import Any, Callable, Iterable, Mapping, cast
from uuid import UUID, uuid4

import attrs

from .._context import current_scheduler
from .._enums import CoalescePolicy, ConflictPolicy, JobOutcome, RunState
from .._events import (
    Event,
    JobReleased,
    ScheduleAdded,
    SchedulerStarted,
    SchedulerStopped,
    ScheduleUpdated,
)
from .._exceptions import (
    JobCancelled,
    JobDeadlineMissed,
    JobLookupError,
    ScheduleLookupError,
)
from .._structures import Job, JobResult, Schedule, Task
from ..abc import DataStore, EventBroker, Trigger
from ..datastores.memory import MemoryDataStore
from ..eventbrokers.local import LocalEventBroker
from ..marshalling import callable_to_ref
from ..workers.sync import Worker

_microsecond_delta = timedelta(microseconds=1)
_zero_timedelta = timedelta()


@attrs.define(eq=False)
class Scheduler:
    """A synchronous scheduler implementation."""

    data_store: DataStore = attrs.field(factory=MemoryDataStore)
    event_broker: EventBroker = attrs.field(factory=LocalEventBroker)
    identity: str = attrs.field(kw_only=True, default=None)
    start_worker: bool = attrs.field(kw_only=True, default=True)
    logger: Logger | None = attrs.field(kw_only=True, default=getLogger(__name__))

    _state: RunState = attrs.field(init=False, default=RunState.stopped)
    _thread: threading.Thread | None = attrs.field(init=False, default=None)
    _wakeup_event: threading.Event = attrs.field(init=False, factory=threading.Event)
    _wakeup_deadline: datetime | None = attrs.field(init=False, default=None)
    _services_initialized: bool = attrs.field(init=False, default=False)
    _exit_stack: ExitStack = attrs.field(init=False, factory=ExitStack)
    _lock: threading.RLock = attrs.field(init=False, factory=threading.RLock)

    def __attrs_post_init__(self) -> None:
        if not self.identity:
            self.identity = f"{platform.node()}-{os.getpid()}-{id(self)}"

    def __enter__(self) -> Scheduler:
        self.start_in_background()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self.stop()
        self._join_thread()

    def _ensure_services_ready(self, exit_stack: ExitStack | None = None) -> None:
        """Ensure that the data store and event broker have been initialized."""
        stack = exit_stack or self._exit_stack
        with self._lock:
            if not self._services_initialized:
                self._services_initialized = True
                self.event_broker.start()
                stack.push(
                    lambda *exc_info: self.event_broker.stop(
                        force=exc_info[0] is not None
                    )
                )

                # Initialize the data store
                self.data_store.start(self.event_broker)
                stack.push(
                    lambda *exc_info: self.data_store.stop(
                        force=exc_info[0] is not None
                    )
                )
                if not exit_stack:
                    atexit.register(self._exit_stack.close)

    def _schedule_added_or_modified(self, event: Event) -> None:
        event_ = cast("ScheduleAdded | ScheduleUpdated", event)
        if not self._wakeup_deadline or (
            event_.next_fire_time and event_.next_fire_time < self._wakeup_deadline
        ):
            self.logger.debug(
                "Detected a %s event – waking up the scheduler", type(event).__name__
            )
            self._wakeup_event.set()

    def _join_thread(self) -> None:
        if self._thread:
            self._thread.join()
            self._thread = None

    @property
    def state(self) -> RunState:
        """The current running state of the scheduler."""
        return self._state

    def add_schedule(
        self,
        func_or_task_id: str | Callable,
        trigger: Trigger,
        *,
        id: str | None = None,
        args: Iterable | None = None,
        kwargs: Mapping[str, Any] | None = None,
        coalesce: CoalescePolicy = CoalescePolicy.latest,
        misfire_grace_time: float | timedelta | None = None,
        max_jitter: float | timedelta | None = None,
        tags: Iterable[str] | None = None,
        conflict_policy: ConflictPolicy = ConflictPolicy.do_nothing,
    ) -> str:
        """
        Schedule a task to be run one or more times in the future.

        :param func_or_task_id: either a callable or an ID of an existing task
            definition
        :param trigger: determines the times when the task should be run
        :param id: an explicit identifier for the schedule (if omitted, a random, UUID
            based ID will be assigned)
        :param args: positional arguments to be passed to the task function
        :param kwargs: keyword arguments to be passed to the task function
        :param coalesce: determines what to do when processing the schedule if multiple
            fire times have become due for this schedule since the last processing
        :param misfire_grace_time: maximum number of seconds the scheduled job's actual
            run time is allowed to be late, compared to the scheduled run time
        :param max_jitter: maximum number of seconds to randomly add to the scheduled
            time for each job created from this schedule
        :param tags: strings that can be used to categorize and filter the schedule and
            its derivative jobs
        :param conflict_policy: determines what to do if a schedule with the same ID
            already exists in the data store
        :return: the ID of the newly added schedule

        """
        self._ensure_services_ready()
        id = id or str(uuid4())
        args = tuple(args or ())
        kwargs = dict(kwargs or {})
        tags = frozenset(tags or ())
        if isinstance(misfire_grace_time, (int, float)):
            misfire_grace_time = timedelta(seconds=misfire_grace_time)

        if callable(func_or_task_id):
            task = Task(id=callable_to_ref(func_or_task_id), func=func_or_task_id)
            self.data_store.add_task(task)
        else:
            task = self.data_store.get_task(func_or_task_id)

        schedule = Schedule(
            id=id,
            task_id=task.id,
            trigger=trigger,
            args=args,
            kwargs=kwargs,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            max_jitter=max_jitter,
            tags=tags,
        )
        schedule.next_fire_time = trigger.next()
        self.data_store.add_schedule(schedule, conflict_policy)
        self.logger.info(
            "Added new schedule (task=%r, trigger=%r); next run time at %s",
            task,
            trigger,
            schedule.next_fire_time,
        )
        return schedule.id

    def get_schedule(self, id: str) -> Schedule:
        """
        Retrieve a schedule from the data store.

        :param id: the unique identifier of the schedule
        :raises ScheduleLookupError: if the schedule could not be found

        """
        self._ensure_services_ready()
        schedules = self.data_store.get_schedules({id})
        if schedules:
            return schedules[0]
        else:
            raise ScheduleLookupError(id)

    def get_schedules(self) -> list[Schedule]:
        """
        Retrieve all schedules from the data store.

        :return: a list of schedules, in an unspecified order

        """
        self._ensure_services_ready()
        return self.data_store.get_schedules()

    def remove_schedule(self, id: str) -> None:
        """
        Remove the given schedule from the data store.

        :param id: the unique identifier of the schedule

        """
        self._ensure_services_ready()
        self.data_store.remove_schedules({id})

    def add_job(
        self,
        func_or_task_id: str | Callable,
        *,
        args: Iterable | None = None,
        kwargs: Mapping[str, Any] | None = None,
        tags: Iterable[str] | None = None,
        result_expiration_time: timedelta | float = 0,
    ) -> UUID:
        """
        Add a job to the data store.

        :param func_or_task_id: either a callable or an ID of an existing task
            definition
        :param args: positional arguments to be passed to the task function
        :param kwargs: keyword arguments to be passed to the task function
        :param tags: strings that can be used to categorize and filter the job
        :param result_expiration_time: the minimum time (as seconds, or timedelta) to
            keep the result of the job available for fetching (the result won't be
            saved at all if that time is 0)
        :return: the ID of the newly created job

        """
        self._ensure_services_ready()
        if callable(func_or_task_id):
            task = Task(id=callable_to_ref(func_or_task_id), func=func_or_task_id)
            self.data_store.add_task(task)
        else:
            task = self.data_store.get_task(func_or_task_id)

        job = Job(
            task_id=task.id,
            args=args or (),
            kwargs=kwargs or {},
            tags=tags or frozenset(),
            result_expiration_time=result_expiration_time,
        )
        self.data_store.add_job(job)
        return job.id

    def get_job_result(self, job_id: UUID, *, wait: bool = True) -> JobResult:
        """
        Retrieve the result of a job.

        :param job_id: the ID of the job
        :param wait: if ``True``, wait until the job has ended (one way or another),
            ``False`` to raise an exception if the result is not yet available
        :raises JobLookupError: if ``wait=False`` and the job result does not exist in
            the data store

        """
        self._ensure_services_ready()
        wait_event = threading.Event()

        def listener(event: JobReleased) -> None:
            if event.job_id == job_id:
                wait_event.set()

        with self.data_store.events.subscribe(listener, {JobReleased}, one_shot=True):
            result = self.data_store.get_job_result(job_id)
            if result:
                return result
            elif not wait:
                raise JobLookupError(job_id)

            wait_event.wait()

        return self.data_store.get_job_result(job_id)

    def run_job(
        self,
        func_or_task_id: str | Callable,
        *,
        args: Iterable | None = None,
        kwargs: Mapping[str, Any] | None = None,
        tags: Iterable[str] | None = (),
    ) -> Any:
        """
        Convenience method to add a job and then return its result.

        If the job raised an exception, that exception will be reraised here.

        :param func_or_task_id: either a callable or an ID of an existing task
            definition
        :param args: positional arguments to be passed to the task function
        :param kwargs: keyword arguments to be passed to the task function
        :param tags: strings that can be used to categorize and filter the job
        :returns: the return value of the task function

        """
        self._ensure_services_ready()
        job_complete_event = threading.Event()

        def listener(event: JobReleased) -> None:
            if event.job_id == job_id:
                job_complete_event.set()

        job_id: UUID | None = None
        with self.data_store.events.subscribe(listener, {JobReleased}):
            job_id = self.add_job(
                func_or_task_id,
                args=args,
                kwargs=kwargs,
                tags=tags,
                result_expiration_time=timedelta(minutes=15),
            )
            job_complete_event.wait()

        result = self.get_job_result(job_id)
        if result.outcome is JobOutcome.success:
            return result.return_value
        elif result.outcome is JobOutcome.error:
            raise result.exception
        elif result.outcome is JobOutcome.missed_start_deadline:
            raise JobDeadlineMissed
        elif result.outcome is JobOutcome.cancelled:
            raise JobCancelled
        else:
            raise RuntimeError(f"Unknown job outcome: {result.outcome}")

    def start_in_background(self) -> None:
        """
        Launch the scheduler in a new thread.

        This method registers :mod:`atexit` hooks to shut down the scheduler and wait
        for the thread to finish.

        :raises RuntimeError: if the scheduler is not in the ``stopped`` state

        """
        with self._lock:
            if self._state is not RunState.stopped:
                raise RuntimeError(
                    f'Cannot start the scheduler when it is in the "{self._state}" '
                    f"state"
                )

            self._state = RunState.starting

        start_future: Future[None] = Future()
        self._thread = threading.Thread(
            target=self._run, args=[start_future], daemon=True
        )
        self._thread.start()
        try:
            start_future.result()
        except BaseException:
            self._thread = None
            raise

        atexit.register(self._join_thread)
        atexit.register(self.stop)

    def stop(self) -> None:
        """
        Signal the scheduler that it should stop processing schedules.

        This method does not wait for the scheduler to actually stop.
        For that, see :meth:`wait_until_stopped`.

        """
        with self._lock:
            if self._state is RunState.started:
                self._state = RunState.stopping
                self._wakeup_event.set()

    def wait_until_stopped(self) -> None:
        """
        Wait until the scheduler is in the "stopped" or "stopping" state.

        If the scheduler is already stopped or in the process of stopping, this method
        returns immediately. Otherwise, it waits until the scheduler posts the
        ``SchedulerStopped`` event.

        """
        with self._lock:
            if self._state in (RunState.stopped, RunState.stopping):
                return

            event = threading.Event()
            sub = self.event_broker.subscribe(
                lambda ev: event.set(), {SchedulerStopped}, one_shot=True
            )

        with sub:
            event.wait()

    def run_until_stopped(self) -> None:
        """
        Run the scheduler (and its internal worker) until it is explicitly stopped.

        This method will only return if :meth:`stop` is called.

        """
        with self._lock:
            if self._state is not RunState.stopped:
                raise RuntimeError(
                    f'Cannot start the scheduler when it is in the "{self._state}" '
                    f"state"
                )

            self._state = RunState.starting

        self._run(None)

    def _run(self, start_future: Future[None] | None) -> None:
        assert self._state is RunState.starting
        with self._exit_stack.pop_all() as exit_stack:
            try:
                self._ensure_services_ready(exit_stack)

                # Wake up the scheduler if the data store emits a significant schedule
                # event
                exit_stack.enter_context(
                    self.data_store.events.subscribe(
                        self._schedule_added_or_modified,
                        {ScheduleAdded, ScheduleUpdated},
                    )
                )

                # Start the built-in worker, if configured to do so
                if self.start_worker:
                    token = current_scheduler.set(self)
                    exit_stack.callback(current_scheduler.reset, token)
                    worker = Worker(
                        self.data_store, self.event_broker, is_internal=True
                    )
                    exit_stack.enter_context(worker)

                # Signal that the scheduler has started
                self._state = RunState.started
                self.event_broker.publish_local(SchedulerStarted())
            except BaseException as exc:
                if start_future:
                    start_future.set_exception(exc)
                    return
                else:
                    raise
            else:
                if start_future:
                    start_future.set_result(None)

            exception: BaseException | None = None
            try:
                while self._state is RunState.started:
                    schedules = self.data_store.acquire_schedules(self.identity, 100)
                    self.logger.debug(
                        "Processing %d schedules retrieved from the data store",
                        len(schedules),
                    )
                    now = datetime.now(timezone.utc)
                    for schedule in schedules:
                        # Calculate a next fire time for the schedule, if possible
                        fire_times = [schedule.next_fire_time]
                        calculate_next = schedule.trigger.next
                        while True:
                            try:
                                fire_time = calculate_next()
                            except Exception:
                                self.logger.exception(
                                    "Error computing next fire time for schedule %r of "
                                    "task %r – removing schedule",
                                    schedule.id,
                                    schedule.task_id,
                                )
                                break

                            # Stop if the calculated fire time is in the future
                            if fire_time is None or fire_time > now:
                                schedule.next_fire_time = fire_time
                                break

                            # Only keep all the fire times if coalesce policy = "all"
                            if schedule.coalesce is CoalescePolicy.all:
                                fire_times.append(fire_time)
                            elif schedule.coalesce is CoalescePolicy.latest:
                                fire_times[0] = fire_time

                        # Add one or more jobs to the job queue
                        max_jitter = (
                            schedule.max_jitter.total_seconds()
                            if schedule.max_jitter
                            else 0
                        )
                        for i, fire_time in enumerate(fire_times):
                            # Calculate a jitter if max_jitter > 0
                            jitter = _zero_timedelta
                            if max_jitter:
                                if i + 1 < len(fire_times):
                                    next_fire_time = fire_times[i + 1]
                                else:
                                    next_fire_time = schedule.next_fire_time

                                if next_fire_time is not None:
                                    # Jitter must never be so high that it would cause
                                    # a fire time to equal or exceed the next fire time
                                    jitter_s = min(
                                        [
                                            max_jitter,
                                            (
                                                next_fire_time
                                                - fire_time
                                                - _microsecond_delta
                                            ).total_seconds(),
                                        ]
                                    )
                                    jitter = timedelta(
                                        seconds=random.uniform(0, jitter_s)
                                    )
                                    fire_time += jitter

                            schedule.last_fire_time = fire_time
                            job = Job(
                                task_id=schedule.task_id,
                                args=schedule.args,
                                kwargs=schedule.kwargs,
                                schedule_id=schedule.id,
                                scheduled_fire_time=fire_time,
                                jitter=jitter,
                                start_deadline=schedule.next_deadline,
                                tags=schedule.tags,
                            )
                            self.data_store.add_job(job)

                    # Update the schedules (and release the scheduler's claim on them)
                    self.data_store.release_schedules(self.identity, schedules)

                    # If we received fewer schedules than the maximum amount, sleep
                    # until the next schedule is due or the scheduler is explicitly
                    # woken up
                    wait_time = None
                    if len(schedules) < 100:
                        self._wakeup_deadline = (
                            self.data_store.get_next_schedule_run_time()
                        )
                        if self._wakeup_deadline:
                            wait_time = (
                                self._wakeup_deadline - datetime.now(timezone.utc)
                            ).total_seconds()
                            self.logger.debug(
                                "Sleeping %.3f seconds until the next fire time (%s)",
                                wait_time,
                                self._wakeup_deadline,
                            )
                        else:
                            self.logger.debug("Waiting for any due schedules to appear")

                        if self._wakeup_event.wait(wait_time):
                            self._wakeup_event = threading.Event()
                    else:
                        self.logger.debug(
                            "Processing more schedules on the next iteration"
                        )
            except BaseException as exc:
                exception = exc
                raise
            finally:
                self._state = RunState.stopped
                if isinstance(exception, Exception):
                    self.logger.exception("Scheduler crashed")
                elif exception:
                    self.logger.info(
                        f"Scheduler stopped due to {exception.__class__.__name__}"
                    )
                else:
                    self.logger.info("Scheduler stopped")

                self.event_broker.publish_local(SchedulerStopped(exception=exception))



"""
Fields represent CronTrigger options which map to :class:`~datetime.datetime` fields.
"""

from __future__ import annotations

import re
from calendar import monthrange
from datetime import datetime
from typing import Any, ClassVar, Sequence

from .expressions import (
    WEEKDAYS,
    AllExpression,
    LastDayOfMonthExpression,
    MonthRangeExpression,
    RangeExpression,
    WeekdayPositionExpression,
    WeekdayRangeExpression,
    get_weekday_index,
)

MIN_VALUES = {
    "year": 1970,
    "month": 1,
    "day": 1,
    "week": 1,
    "day_of_week": 0,
    "hour": 0,
    "minute": 0,
    "second": 0,
}
MAX_VALUES = {
    "year": 9999,
    "month": 12,
    "day": 31,
    "week": 53,
    "day_of_week": 7,
    "hour": 23,
    "minute": 59,
    "second": 59,
}
DEFAULT_VALUES = {
    "year": "*",
    "month": 1,
    "day": 1,
    "week": "*",
    "day_of_week": "*",
    "hour": 0,
    "minute": 0,
    "second": 0,
}
SEPARATOR = re.compile(" *, *")


class BaseField:
    __slots__ = "name", "expressions"

    real: ClassVar[bool] = True
    compilers: ClassVar[Any] = (AllExpression, RangeExpression)

    def __init_subclass__(cls, real: bool = True, extra_compilers: Sequence = ()):
        cls.real = real
        if extra_compilers:
            cls.compilers += extra_compilers

    def __init__(self, name: str, exprs: int | str):
        self.name = name
        self.expressions: list = []
        for expr in SEPARATOR.split(str(exprs).strip()):
            self.append_expression(expr)

    def get_min(self, dateval: datetime) -> int:
        return MIN_VALUES[self.name]

    def get_max(self, dateval: datetime) -> int:
        return MAX_VALUES[self.name]

    def get_value(self, dateval: datetime) -> int:
        return getattr(dateval, self.name)

    def get_next_value(self, dateval: datetime) -> int | None:
        smallest = None
        for expr in self.expressions:
            value = expr.get_next_value(dateval, self)
            if smallest is None or (value is not None and value < smallest):
                smallest = value

        return smallest

    def append_expression(self, expr: str) -> None:
        for compiler in self.compilers:
            match = compiler.value_re.match(expr)
            if match:
                compiled_expr = compiler(**match.groupdict())

                try:
                    compiled_expr.validate_range(
                        self.name, MIN_VALUES[self.name], MAX_VALUES[self.name]
                    )
                except ValueError as exc:
                    raise ValueError(
                        f"Error validating expression {expr!r}: {exc}"
                    ) from exc

                self.expressions.append(compiled_expr)
                return

        raise ValueError(f"Unrecognized expression {expr!r} for field {self.name!r}")

    def __str__(self):
        expr_strings = (str(e) for e in self.expressions)
        return ",".join(expr_strings)


class WeekField(BaseField, real=False):
    __slots__ = ()

    def get_value(self, dateval: datetime) -> int:
        return dateval.isocalendar()[1]


class DayOfMonthField(
    BaseField, extra_compilers=(WeekdayPositionExpression, LastDayOfMonthExpression)
):
    __slots__ = ()

    def get_max(self, dateval: datetime) -> int:
        return monthrange(dateval.year, dateval.month)[1]


class DayOfWeekField(BaseField, real=False, extra_compilers=(WeekdayRangeExpression,)):
    __slots__ = ()

    def append_expression(self, expr: str) -> None:
        # Convert numeric weekday expressions into textual ones
        match = RangeExpression.value_re.match(expr)
        if match:
            groups = match.groups()
            first = int(groups[0]) - 1
            first = 6 if first < 0 else first
            if groups[1]:
                last = int(groups[1]) - 1
                last = 6 if last < 0 else last
            else:
                last = first

            expr = f"{WEEKDAYS[first]}-{WEEKDAYS[last]}"

        # For expressions like Sun-Tue or Sat-Mon, add two expressions that together
        # cover the expected weekdays
        match = WeekdayRangeExpression.value_re.match(expr)
        if match and match.groups()[1]:
            groups = match.groups()
            first_index = get_weekday_index(groups[0])
            last_index = get_weekday_index(groups[1])
            if first_index > last_index:
                super().append_expression(f"{WEEKDAYS[0]}-{groups[1]}")
                super().append_expression(f"{groups[0]}-{WEEKDAYS[-1]}")
                return

        super().append_expression(expr)

    def get_value(self, dateval: datetime) -> int:
        return dateval.weekday()


class MonthField(BaseField, extra_compilers=(MonthRangeExpression,)):
    __slots__ = ()



from __future__ import annotations

from collections import defaultdict
from datetime import datetime, timedelta, timezone
from logging import Logger, getLogger
from typing import Any, Iterable
from uuid import UUID

import attrs
import tenacity
from sqlalchemy import (
    JSON,
    TIMESTAMP,
    BigInteger,
    Column,
    Enum,
    Integer,
    LargeBinary,
    MetaData,
    Table,
    TypeDecorator,
    Unicode,
    and_,
    bindparam,
    or_,
    select,
)
from sqlalchemy.engine import URL, Dialect, Result
from sqlalchemy.exc import CompileError, IntegrityError, OperationalError
from sqlalchemy.future import Engine, create_engine
from sqlalchemy.sql.ddl import DropTable
from sqlalchemy.sql.elements import BindParameter, literal

from .._enums import CoalescePolicy, ConflictPolicy, JobOutcome
from .._events import (
    Event,
    JobAcquired,
    JobAdded,
    JobDeserializationFailed,
    ScheduleAdded,
    ScheduleDeserializationFailed,
    ScheduleRemoved,
    ScheduleUpdated,
    TaskAdded,
    TaskRemoved,
    TaskUpdated,
)
from .._exceptions import ConflictingIdError, SerializationError, TaskLookupError
from .._structures import Job, JobResult, RetrySettings, Schedule, Task
from ..abc import EventBroker, Serializer
from ..marshalling import callable_to_ref
from ..serializers.pickle import PickleSerializer
from .base import BaseDataStore


class EmulatedUUID(TypeDecorator):
    impl = Unicode(32)
    cache_ok = True

    def process_bind_param(self, value, dialect: Dialect) -> Any:
        return value.hex if value is not None else None

    def process_result_value(self, value: Any, dialect: Dialect):
        return UUID(value) if value else None


class EmulatedTimestampTZ(TypeDecorator):
    impl = Unicode(32)
    cache_ok = True

    def process_bind_param(self, value, dialect: Dialect) -> Any:
        return value.isoformat() if value is not None else None

    def process_result_value(self, value: Any, dialect: Dialect):
        return datetime.fromisoformat(value) if value is not None else None


class EmulatedInterval(TypeDecorator):
    impl = BigInteger()
    cache_ok = True

    def process_bind_param(self, value, dialect: Dialect) -> Any:
        return value.total_seconds() if value is not None else None

    def process_result_value(self, value: Any, dialect: Dialect):
        return timedelta(seconds=value) if value is not None else None


@attrs.define(kw_only=True, eq=False)
class _BaseSQLAlchemyDataStore:
    schema: str | None = attrs.field(default=None)
    serializer: Serializer = attrs.field(factory=PickleSerializer)
    lock_expiration_delay: float = attrs.field(default=30)
    max_poll_time: float | None = attrs.field(default=1)
    max_idle_time: float = attrs.field(default=60)
    retry_settings: RetrySettings = attrs.field(default=RetrySettings())
    start_from_scratch: bool = attrs.field(default=False)

    _logger: Logger = attrs.field(init=False, factory=lambda: getLogger(__name__))

    def __attrs_post_init__(self) -> None:
        # Generate the table definitions
        self._metadata = self.get_table_definitions()
        self.t_metadata = self._metadata.tables["metadata"]
        self.t_tasks = self._metadata.tables["tasks"]
        self.t_schedules = self._metadata.tables["schedules"]
        self.t_jobs = self._metadata.tables["jobs"]
        self.t_job_results = self._metadata.tables["job_results"]

        # Find out if the dialect supports UPDATE...RETURNING
        update = self.t_jobs.update().returning(self.t_jobs.c.id)
        try:
            update.compile(bind=self.engine)
        except CompileError:
            self._supports_update_returning = False
        else:
            self._supports_update_returning = True

    def _after_attempt(self, retry_state: tenacity.RetryCallState) -> None:
        self._logger.warning(
            "Temporary data store error (attempt %d): %s",
            retry_state.attempt_number,
            retry_state.outcome.exception(),
        )

    def get_table_definitions(self) -> MetaData:
        if self.engine.dialect.name == "postgresql":
            from sqlalchemy.dialects import postgresql

            timestamp_type = TIMESTAMP(timezone=True)
            job_id_type = postgresql.UUID(as_uuid=True)
            interval_type = postgresql.INTERVAL(precision=6)
            tags_type = postgresql.ARRAY(Unicode)
        else:
            timestamp_type = EmulatedTimestampTZ
            job_id_type = EmulatedUUID
            interval_type = EmulatedInterval
            tags_type = JSON

        metadata = MetaData()
        Table("metadata", metadata, Column("schema_version", Integer, nullable=False))
        Table(
            "tasks",
            metadata,
            Column("id", Unicode(500), primary_key=True),
            Column("func", Unicode(500), nullable=False),
            Column("state", LargeBinary),
            Column("max_running_jobs", Integer),
            Column("misfire_grace_time", interval_type),
            Column("running_jobs", Integer, nullable=False, server_default=literal(0)),
        )
        Table(
            "schedules",
            metadata,
            Column("id", Unicode(500), primary_key=True),
            Column("task_id", Unicode(500), nullable=False, index=True),
            Column("trigger", LargeBinary),
            Column("args", LargeBinary),
            Column("kwargs", LargeBinary),
            Column("coalesce", Enum(CoalescePolicy), nullable=False),
            Column("misfire_grace_time", interval_type),
            Column("max_jitter", interval_type),
            Column("tags", tags_type, nullable=False),
            Column("next_fire_time", timestamp_type, index=True),
            Column("last_fire_time", timestamp_type),
            Column("acquired_by", Unicode(500)),
            Column("acquired_until", timestamp_type),
        )
        Table(
            "jobs",
            metadata,
            Column("id", job_id_type, primary_key=True),
            Column("task_id", Unicode(500), nullable=False, index=True),
            Column("args", LargeBinary, nullable=False),
            Column("kwargs", LargeBinary, nullable=False),
            Column("schedule_id", Unicode(500)),
            Column("scheduled_fire_time", timestamp_type),
            Column("jitter", interval_type),
            Column("start_deadline", timestamp_type),
            Column("result_expiration_time", interval_type),
            Column("tags", tags_type, nullable=False),
            Column("created_at", timestamp_type, nullable=False),
            Column("started_at", timestamp_type),
            Column("acquired_by", Unicode(500)),
            Column("acquired_until", timestamp_type),
        )
        Table(
            "job_results",
            metadata,
            Column("job_id", job_id_type, primary_key=True),
            Column("outcome", Enum(JobOutcome), nullable=False),
            Column("finished_at", timestamp_type, index=True),
            Column("expires_at", timestamp_type, nullable=False, index=True),
            Column("exception", LargeBinary),
            Column("return_value", LargeBinary),
        )
        return metadata

    def _deserialize_schedules(self, result: Result) -> list[Schedule]:
        schedules: list[Schedule] = []
        for row in result:
            try:
                schedules.append(Schedule.unmarshal(self.serializer, row._asdict()))
            except SerializationError as exc:
                self._events.publish(
                    ScheduleDeserializationFailed(schedule_id=row["id"], exception=exc)
                )

        return schedules

    def _deserialize_jobs(self, result: Result) -> list[Job]:
        jobs: list[Job] = []
        for row in result:
            try:
                jobs.append(Job.unmarshal(self.serializer, row._asdict()))
            except SerializationError as exc:
                self._events.publish(
                    JobDeserializationFailed(job_id=row["id"], exception=exc)
                )

        return jobs


@attrs.define(eq=False)
class SQLAlchemyDataStore(_BaseSQLAlchemyDataStore, BaseDataStore):
    """
    Uses a relational database to store data.

    When started, this data store creates the appropriate tables on the given database
    if they're not already present.

    Operations are retried (in accordance to ``retry_settings``) when an operation
    raises :exc:`sqlalchemy.OperationalError`.

    This store has been tested to work with PostgreSQL (psycopg2 driver), MySQL
    (pymysql driver) and SQLite.

    :param engine: a (synchronous) SQLAlchemy engine
    :param schema: a database schema name to use, if not the default
    :param serializer: the serializer used to (de)serialize tasks, schedules and jobs
        for storage
    :param lock_expiration_delay: maximum amount of time (in seconds) that a scheduler
        or worker can keep a lock on a schedule or task
    :param retry_settings: Tenacity settings for retrying operations in case of a
        database connecitivty problem
    :param start_from_scratch: erase all existing data during startup (useful for test
        suites)
    """

    engine: Engine

    @classmethod
    def from_url(cls, url: str | URL, **kwargs) -> SQLAlchemyDataStore:
        """
        Create a new SQLAlchemy data store.

        :param url: an SQLAlchemy URL to pass to :func:`~sqlalchemy.create_engine`
        :param kwargs: keyword arguments to pass to the initializer of this class
        :return: the newly created data store

        """
        engine = create_engine(url)
        return cls(engine, **kwargs)

    def _retry(self) -> tenacity.Retrying:
        return tenacity.Retrying(
            stop=self.retry_settings.stop,
            wait=self.retry_settings.wait,
            retry=tenacity.retry_if_exception_type(OperationalError),
            after=self._after_attempt,
            reraise=True,
        )

    def start(self, event_broker: EventBroker) -> None:
        super().start(event_broker)

        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                if self.start_from_scratch:
                    for table in self._metadata.sorted_tables:
                        conn.execute(DropTable(table, if_exists=True))

                self._metadata.create_all(conn)
                query = select(self.t_metadata.c.schema_version)
                result = conn.execute(query)
                version = result.scalar()
                if version is None:
                    conn.execute(self.t_metadata.insert(values={"schema_version": 1}))
                elif version > 1:
                    raise RuntimeError(
                        f"Unexpected schema version ({version}); "
                        f"only version 1 is supported by this version of APScheduler"
                    )

    def add_task(self, task: Task) -> None:
        insert = self.t_tasks.insert().values(
            id=task.id,
            func=callable_to_ref(task.func),
            max_running_jobs=task.max_running_jobs,
            misfire_grace_time=task.misfire_grace_time,
        )
        try:
            for attempt in self._retry():
                with attempt, self.engine.begin() as conn:
                    conn.execute(insert)
        except IntegrityError:
            update = (
                self.t_tasks.update()
                .values(
                    func=callable_to_ref(task.func),
                    max_running_jobs=task.max_running_jobs,
                    misfire_grace_time=task.misfire_grace_time,
                )
                .where(self.t_tasks.c.id == task.id)
            )
            for attempt in self._retry():
                with attempt, self.engine.begin() as conn:
                    conn.execute(update)
                    self._events.publish(TaskUpdated(task_id=task.id))
        else:
            self._events.publish(TaskAdded(task_id=task.id))

    def remove_task(self, task_id: str) -> None:
        delete = self.t_tasks.delete().where(self.t_tasks.c.id == task_id)
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                result = conn.execute(delete)
                if result.rowcount == 0:
                    raise TaskLookupError(task_id)
                else:
                    self._events.publish(TaskRemoved(task_id=task_id))

    def get_task(self, task_id: str) -> Task:
        query = select(
            [
                self.t_tasks.c.id,
                self.t_tasks.c.func,
                self.t_tasks.c.max_running_jobs,
                self.t_tasks.c.state,
                self.t_tasks.c.misfire_grace_time,
            ]
        ).where(self.t_tasks.c.id == task_id)
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                result = conn.execute(query)
                row = result.first()

        if row:
            return Task.unmarshal(self.serializer, row._asdict())
        else:
            raise TaskLookupError(task_id)

    def get_tasks(self) -> list[Task]:
        query = select(
            [
                self.t_tasks.c.id,
                self.t_tasks.c.func,
                self.t_tasks.c.max_running_jobs,
                self.t_tasks.c.state,
                self.t_tasks.c.misfire_grace_time,
            ]
        ).order_by(self.t_tasks.c.id)
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                result = conn.execute(query)
                tasks = [
                    Task.unmarshal(self.serializer, row._asdict()) for row in result
                ]
                return tasks

    def add_schedule(self, schedule: Schedule, conflict_policy: ConflictPolicy) -> None:
        event: Event
        values = schedule.marshal(self.serializer)
        insert = self.t_schedules.insert().values(**values)
        try:
            for attempt in self._retry():
                with attempt, self.engine.begin() as conn:
                    conn.execute(insert)
        except IntegrityError:
            if conflict_policy is ConflictPolicy.exception:
                raise ConflictingIdError(schedule.id) from None
            elif conflict_policy is ConflictPolicy.replace:
                del values["id"]
                update = (
                    self.t_schedules.update()
                    .where(self.t_schedules.c.id == schedule.id)
                    .values(**values)
                )
                for attempt in self._retry():
                    with attempt, self.engine.begin() as conn:
                        conn.execute(update)

                event = ScheduleUpdated(
                    schedule_id=schedule.id, next_fire_time=schedule.next_fire_time
                )
                self._events.publish(event)
        else:
            event = ScheduleAdded(
                schedule_id=schedule.id, next_fire_time=schedule.next_fire_time
            )
            self._events.publish(event)

    def remove_schedules(self, ids: Iterable[str]) -> None:
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                delete = self.t_schedules.delete().where(self.t_schedules.c.id.in_(ids))
                if self._supports_update_returning:
                    delete = delete.returning(self.t_schedules.c.id)
                    removed_ids: Iterable[str] = [
                        row[0] for row in conn.execute(delete)
                    ]
                else:
                    # TODO: actually check which rows were deleted?
                    conn.execute(delete)
                    removed_ids = ids

        for schedule_id in removed_ids:
            self._events.publish(ScheduleRemoved(schedule_id=schedule_id))

    def get_schedules(self, ids: set[str] | None = None) -> list[Schedule]:
        query = self.t_schedules.select().order_by(self.t_schedules.c.id)
        if ids:
            query = query.where(self.t_schedules.c.id.in_(ids))

        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                result = conn.execute(query)
                return self._deserialize_schedules(result)

    def acquire_schedules(self, scheduler_id: str, limit: int) -> list[Schedule]:
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                now = datetime.now(timezone.utc)
                acquired_until = now + timedelta(seconds=self.lock_expiration_delay)
                schedules_cte = (
                    select(self.t_schedules.c.id)
                    .where(
                        and_(
                            self.t_schedules.c.next_fire_time.isnot(None),
                            self.t_schedules.c.next_fire_time <= now,
                            or_(
                                self.t_schedules.c.acquired_until.is_(None),
                                self.t_schedules.c.acquired_until < now,
                            ),
                        )
                    )
                    .order_by(self.t_schedules.c.next_fire_time)
                    .limit(limit)
                    .with_for_update(skip_locked=True)
                    .cte()
                )
                subselect = select([schedules_cte.c.id])
                update = (
                    self.t_schedules.update()
                    .where(self.t_schedules.c.id.in_(subselect))
                    .values(acquired_by=scheduler_id, acquired_until=acquired_until)
                )
                if self._supports_update_returning:
                    update = update.returning(*self.t_schedules.columns)
                    result = conn.execute(update)
                else:
                    conn.execute(update)
                    query = self.t_schedules.select().where(
                        and_(self.t_schedules.c.acquired_by == scheduler_id)
                    )
                    result = conn.execute(query)

                schedules = self._deserialize_schedules(result)

        return schedules

    def release_schedules(self, scheduler_id: str, schedules: list[Schedule]) -> None:
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                update_events: list[ScheduleUpdated] = []
                finished_schedule_ids: list[str] = []
                update_args: list[dict[str, Any]] = []
                for schedule in schedules:
                    if schedule.next_fire_time is not None:
                        try:
                            serialized_trigger = self.serializer.serialize(
                                schedule.trigger
                            )
                        except SerializationError:
                            self._logger.exception(
                                "Error serializing trigger for schedule %r – "
                                "removing from data store",
                                schedule.id,
                            )
                            finished_schedule_ids.append(schedule.id)
                            continue

                        update_args.append(
                            {
                                "p_id": schedule.id,
                                "p_trigger": serialized_trigger,
                                "p_next_fire_time": schedule.next_fire_time,
                            }
                        )
                    else:
                        finished_schedule_ids.append(schedule.id)

                # Update schedules that have a next fire time
                if update_args:
                    p_id: BindParameter = bindparam("p_id")
                    p_trigger: BindParameter = bindparam("p_trigger")
                    p_next_fire_time: BindParameter = bindparam("p_next_fire_time")
                    update = (
                        self.t_schedules.update()
                        .where(
                            and_(
                                self.t_schedules.c.id == p_id,
                                self.t_schedules.c.acquired_by == scheduler_id,
                            )
                        )
                        .values(
                            trigger=p_trigger,
                            next_fire_time=p_next_fire_time,
                            acquired_by=None,
                            acquired_until=None,
                        )
                    )
                    next_fire_times = {
                        arg["p_id"]: arg["p_next_fire_time"] for arg in update_args
                    }
                    # TODO: actually check which rows were updated?
                    conn.execute(update, update_args)
                    updated_ids = list(next_fire_times)

                    for schedule_id in updated_ids:
                        event = ScheduleUpdated(
                            schedule_id=schedule_id,
                            next_fire_time=next_fire_times[schedule_id],
                        )
                        update_events.append(event)

                # Remove schedules that have no next fire time or failed to serialize
                if finished_schedule_ids:
                    delete = self.t_schedules.delete().where(
                        self.t_schedules.c.id.in_(finished_schedule_ids)
                    )
                    conn.execute(delete)

        for event in update_events:
            self._events.publish(event)

        for schedule_id in finished_schedule_ids:
            self._events.publish(ScheduleRemoved(schedule_id=schedule_id))

    def get_next_schedule_run_time(self) -> datetime | None:
        query = (
            select(self.t_schedules.c.next_fire_time)
            .where(self.t_schedules.c.next_fire_time.isnot(None))
            .order_by(self.t_schedules.c.next_fire_time)
            .limit(1)
        )
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                result = conn.execute(query)
                return result.scalar()

    def add_job(self, job: Job) -> None:
        marshalled = job.marshal(self.serializer)
        insert = self.t_jobs.insert().values(**marshalled)
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                conn.execute(insert)

        event = JobAdded(
            job_id=job.id,
            task_id=job.task_id,
            schedule_id=job.schedule_id,
            tags=job.tags,
        )
        self._events.publish(event)

    def get_jobs(self, ids: Iterable[UUID] | None = None) -> list[Job]:
        query = self.t_jobs.select().order_by(self.t_jobs.c.id)
        if ids:
            job_ids = [job_id for job_id in ids]
            query = query.where(self.t_jobs.c.id.in_(job_ids))

        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                result = conn.execute(query)
                return self._deserialize_jobs(result)

    def acquire_jobs(self, worker_id: str, limit: int | None = None) -> list[Job]:
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                now = datetime.now(timezone.utc)
                acquired_until = now + timedelta(seconds=self.lock_expiration_delay)
                query = (
                    self.t_jobs.select()
                    .join(self.t_tasks, self.t_tasks.c.id == self.t_jobs.c.task_id)
                    .where(
                        or_(
                            self.t_jobs.c.acquired_until.is_(None),
                            self.t_jobs.c.acquired_until < now,
                        )
                    )
                    .order_by(self.t_jobs.c.created_at)
                    .with_for_update(skip_locked=True)
                    .limit(limit)
                )

                result = conn.execute(query)
                if not result:
                    return []

                # Mark the jobs as acquired by this worker
                jobs = self._deserialize_jobs(result)
                task_ids: set[str] = {job.task_id for job in jobs}

                # Retrieve the limits
                query = select(
                    [
                        self.t_tasks.c.id,
                        self.t_tasks.c.max_running_jobs - self.t_tasks.c.running_jobs,
                    ]
                ).where(
                    self.t_tasks.c.max_running_jobs.isnot(None),
                    self.t_tasks.c.id.in_(task_ids),
                )
                result = conn.execute(query)
                job_slots_left = dict(result.fetchall())

                # Filter out jobs that don't have free slots
                acquired_jobs: list[Job] = []
                increments: dict[str, int] = defaultdict(lambda: 0)
                for job in jobs:
                    # Don't acquire the job if there are no free slots left
                    slots_left = job_slots_left.get(job.task_id)
                    if slots_left == 0:
                        continue
                    elif slots_left is not None:
                        job_slots_left[job.task_id] -= 1

                    acquired_jobs.append(job)
                    increments[job.task_id] += 1

                if acquired_jobs:
                    # Mark the acquired jobs as acquired by this worker
                    acquired_job_ids = [job.id for job in acquired_jobs]
                    update = (
                        self.t_jobs.update()
                        .values(acquired_by=worker_id, acquired_until=acquired_until)
                        .where(self.t_jobs.c.id.in_(acquired_job_ids))
                    )
                    conn.execute(update)

                    # Increment the running job counters on each task
                    p_id: BindParameter = bindparam("p_id")
                    p_increment: BindParameter = bindparam("p_increment")
                    params = [
                        {"p_id": task_id, "p_increment": increment}
                        for task_id, increment in increments.items()
                    ]
                    update = (
                        self.t_tasks.update()
                        .values(running_jobs=self.t_tasks.c.running_jobs + p_increment)
                        .where(self.t_tasks.c.id == p_id)
                    )
                    conn.execute(update, params)

        # Publish the appropriate events
        for job in acquired_jobs:
            self._events.publish(JobAcquired(job_id=job.id, worker_id=worker_id))

        return acquired_jobs

    def release_job(self, worker_id: str, task_id: str, result: JobResult) -> None:
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                # Insert the job result
                if result.expires_at > result.finished_at:
                    marshalled = result.marshal(self.serializer)
                    insert = self.t_job_results.insert().values(**marshalled)
                    conn.execute(insert)

                # Decrement the running jobs counter
                update = (
                    self.t_tasks.update()
                    .values(running_jobs=self.t_tasks.c.running_jobs - 1)
                    .where(self.t_tasks.c.id == task_id)
                )
                conn.execute(update)

                # Delete the job
                delete = self.t_jobs.delete().where(self.t_jobs.c.id == result.job_id)
                conn.execute(delete)

    def get_job_result(self, job_id: UUID) -> JobResult | None:
        for attempt in self._retry():
            with attempt, self.engine.begin() as conn:
                # Retrieve the result
                query = self.t_job_results.select().where(
                    self.t_job_results.c.job_id == job_id
                )
                row = conn.execute(query).first()

                # Delete the result
                delete = self.t_job_results.delete().where(
                    self.t_job_results.c.job_id == job_id
                )
                conn.execute(delete)

                return (
                    JobResult.unmarshal(self.serializer, row._asdict()) if row else None
                )



from __future__ import annotations

from enum import Enum, auto


class RunState(Enum):
    """
    Used to track the running state of schedulers and workers.

    Values:

    * ``starting``: not running yet, but in the process of starting
    * ``started``: running
    * ``stopping``: still running but in the process of shutting down
    * ``stopped``: not running
    """

    starting = auto()
    started = auto()
    stopping = auto()
    stopped = auto()


class JobOutcome(Enum):
    """
    Used to indicate how the execution of a job ended.

    Values:

    * ``success``: the job completed successfully
    * ``error``: the job raised an exception
    * ``missed_start_deadline``: the job's execution was delayed enough for it to miss
      its deadline
    * ``cancelled``: the job's execution was cancelled
    """

    success = auto()
    error = auto()
    missed_start_deadline = auto()
    cancelled = auto()


class ConflictPolicy(Enum):
    """
    Used to indicate what to do when trying to add a schedule whose ID conflicts with an
    existing schedule.

    Values:

    * ``replace``: replace the existing schedule with a new one
    * ``do_nothing``: keep the existing schedule as-is and drop the new schedule
    * ``exception``: raise an exception if a conflict is detected
    """

    replace = auto()
    do_nothing = auto()
    exception = auto()


class CoalescePolicy(Enum):
    """
    Used to indicate how to queue jobs for a schedule that has accumulated multiple
    run times since the last scheduler iteration.

    Values:

    * ``earliest``: run once, with the earliest fire time
    * ``latest``: run once, with the latest fire time
    * ``all``: submit one job for every accumulated fire time
    """

    earliest = auto()
    latest = auto()
    all = auto()



from __future__ import annotations

import os
import platform
from contextlib import AsyncExitStack
from datetime import datetime, timezone
from inspect import isawaitable
from logging import Logger, getLogger
from types import TracebackType
from typing import Callable
from uuid import UUID

import anyio
import attrs
from anyio import (
    TASK_STATUS_IGNORED,
    create_task_group,
    get_cancelled_exc_class,
    move_on_after,
)
from anyio.abc import CancelScope, TaskGroup

from .._context import current_job, current_worker
from .._converters import as_async_datastore, as_async_eventbroker
from .._enums import JobOutcome, RunState
from .._events import JobAdded, JobReleased, WorkerStarted, WorkerStopped
from .._structures import Job, JobInfo, JobResult
from .._validators import positive_integer
from ..abc import AsyncDataStore, AsyncEventBroker
from ..eventbrokers.async_local import LocalAsyncEventBroker


@attrs.define(eq=False)
class AsyncWorker:
    """Runs jobs locally in a task group."""

    data_store: AsyncDataStore = attrs.field(converter=as_async_datastore)
    event_broker: AsyncEventBroker = attrs.field(
        converter=as_async_eventbroker, factory=LocalAsyncEventBroker
    )
    max_concurrent_jobs: int = attrs.field(
        kw_only=True, validator=positive_integer, default=100
    )
    identity: str = attrs.field(kw_only=True, default=None)
    logger: Logger | None = attrs.field(kw_only=True, default=getLogger(__name__))
    # True if a scheduler owns this worker
    _is_internal: bool = attrs.field(kw_only=True, default=False)

    _state: RunState = attrs.field(init=False, default=RunState.stopped)
    _wakeup_event: anyio.Event = attrs.field(init=False)
    _task_group: TaskGroup = attrs.field(init=False)
    _acquired_jobs: set[Job] = attrs.field(init=False, factory=set)
    _running_jobs: set[UUID] = attrs.field(init=False, factory=set)

    def __attrs_post_init__(self) -> None:
        if not self.identity:
            self.identity = f"{platform.node()}-{os.getpid()}-{id(self)}"

    async def __aenter__(self) -> AsyncWorker:
        self._task_group = create_task_group()
        await self._task_group.__aenter__()
        await self._task_group.start(self.run_until_stopped)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self._state = RunState.stopping
        self._wakeup_event.set()
        await self._task_group.__aexit__(exc_type, exc_val, exc_tb)
        del self._task_group
        del self._wakeup_event

    @property
    def state(self) -> RunState:
        """The current running state of the worker."""
        return self._state

    async def run_until_stopped(self, *, task_status=TASK_STATUS_IGNORED) -> None:
        """Run the worker until it is explicitly stopped."""
        if self._state is not RunState.stopped:
            raise RuntimeError(
                f'Cannot start the worker when it is in the "{self._state}" ' f"state"
            )

        self._state = RunState.starting
        self._wakeup_event = anyio.Event()
        async with AsyncExitStack() as exit_stack:
            if not self._is_internal:
                # Initialize the event broker
                await self.event_broker.start()
                exit_stack.push_async_exit(
                    lambda *exc_info: self.event_broker.stop(
                        force=exc_info[0] is not None
                    )
                )

                # Initialize the data store
                await self.data_store.start(self.event_broker)
                exit_stack.push_async_exit(
                    lambda *exc_info: self.data_store.stop(
                        force=exc_info[0] is not None
                    )
                )

            # Set the current worker
            token = current_worker.set(self)
            exit_stack.callback(current_worker.reset, token)

            # Wake up the worker if the data store emits a significant job event
            self.event_broker.subscribe(
                lambda event: self._wakeup_event.set(), {JobAdded}
            )

            # Signal that the worker has started
            self._state = RunState.started
            task_status.started()
            exception: BaseException | None = None
            try:
                await self.event_broker.publish_local(WorkerStarted())

                async with create_task_group() as tg:
                    while self._state is RunState.started:
                        limit = self.max_concurrent_jobs - len(self._running_jobs)
                        jobs = await self.data_store.acquire_jobs(self.identity, limit)
                        for job in jobs:
                            task = await self.data_store.get_task(job.task_id)
                            self._running_jobs.add(job.id)
                            tg.start_soon(self._run_job, job, task.func)

                        await self._wakeup_event.wait()
                        self._wakeup_event = anyio.Event()
            except get_cancelled_exc_class():
                pass
            except BaseException as exc:
                self.logger.exception("Worker crashed")
                exception = exc
            else:
                self.logger.info("Worker stopped")
            finally:
                self._state = RunState.stopped
                with move_on_after(3, shield=True):
                    await self.event_broker.publish_local(
                        WorkerStopped(exception=exception)
                    )

    async def stop(self, *, force: bool = False) -> None:
        """
        Signal the worker that it should stop running jobs.

        This method does not wait for the worker to actually stop.

        """
        if self._state in (RunState.starting, RunState.started):
            self._state = RunState.stopping
            event = anyio.Event()
            self.event_broker.subscribe(
                lambda ev: event.set(), {WorkerStopped}, one_shot=True
            )
            if force:
                self._task_group.cancel_scope.cancel()
            else:
                self._wakeup_event.set()

            await event.wait()

    async def _run_job(self, job: Job, func: Callable) -> None:
        try:
            # Check if the job started before the deadline
            start_time = datetime.now(timezone.utc)
            if job.start_deadline is not None and start_time > job.start_deadline:
                result = JobResult.from_job(
                    job,
                    outcome=JobOutcome.missed_start_deadline,
                    finished_at=start_time,
                )
                await self.data_store.release_job(self.identity, job.task_id, result)
                await self.event_broker.publish(
                    JobReleased.from_result(result, self.identity)
                )
                return

            token = current_job.set(JobInfo.from_job(job))
            try:
                retval = func(*job.args, **job.kwargs)
                if isawaitable(retval):
                    retval = await retval
            except get_cancelled_exc_class():
                self.logger.info("Job %s was cancelled", job.id)
                with CancelScope(shield=True):
                    result = JobResult.from_job(
                        job,
                        outcome=JobOutcome.cancelled,
                    )
                    await self.data_store.release_job(
                        self.identity, job.task_id, result
                    )
                    await self.event_broker.publish(
                        JobReleased.from_result(result, self.identity)
                    )
            except BaseException as exc:
                if isinstance(exc, Exception):
                    self.logger.exception("Job %s raised an exception", job.id)
                else:
                    self.logger.error(
                        "Job %s was aborted due to %s", job.id, exc.__class__.__name__
                    )

                result = JobResult.from_job(
                    job,
                    JobOutcome.error,
                    exception=exc,
                )
                await self.data_store.release_job(
                    self.identity,
                    job.task_id,
                    result,
                )
                await self.event_broker.publish(
                    JobReleased.from_result(result, self.identity)
                )
                if not isinstance(exc, Exception):
                    raise
            else:
                self.logger.info("Job %s completed successfully", job.id)
                result = JobResult.from_job(
                    job,
                    JobOutcome.success,
                    return_value=retval,
                )
                await self.data_store.release_job(self.identity, job.task_id, result)
                await self.event_broker.publish(
                    JobReleased.from_result(result, self.identity)
                )
            finally:
                current_job.reset(token)
        finally:
            self._running_jobs.remove(job.id)



from __future__ import annotations

from concurrent.futures import Future
from threading import Thread

import attrs
from redis import ConnectionPool, Redis

from .._events import Event
from ..abc import Serializer
from ..serializers.json import JSONSerializer
from .base import DistributedEventBrokerMixin
from .local import LocalEventBroker


@attrs.define(eq=False)
class RedisEventBroker(LocalEventBroker, DistributedEventBrokerMixin):
    """
    An event broker that uses a Redis server to broadcast events.

    Requires the redis_ library to be installed.

    .. _redis: https://pypi.org/project/redis/

    :param client: a (synchronous) Redis client
    :param serializer: the serializer used to (de)serialize events for transport
    :param channel: channel on which to send the messages
    :param message_poll_interval: interval on which to poll for new messages (higher
        values mean slower reaction time but less CPU use)
    """

    client: Redis
    serializer: Serializer = attrs.field(factory=JSONSerializer)
    channel: str = attrs.field(kw_only=True, default="apscheduler")
    message_poll_interval: float = attrs.field(kw_only=True, default=0.05)
    _stopped: bool = attrs.field(init=False, default=True)
    _ready_future: Future[None] = attrs.field(init=False)
    _thread: Thread = attrs.field(init=False)

    @classmethod
    def from_url(cls, url: str, **kwargs) -> RedisEventBroker:
        """
        Create a new event broker from a URL.

        :param url: a Redis URL (```redis://...```)
        :param kwargs: keyword arguments to pass to the initializer of this class
        :return: the newly created event broker

        """
        pool = ConnectionPool.from_url(url, **kwargs)
        client = Redis(connection_pool=pool)
        return cls(client)

    def start(self) -> None:
        self._stopped = False
        self._ready_future = Future()
        self._thread = Thread(
            target=self._listen_messages, daemon=True, name="Redis subscriber"
        )
        self._thread.start()
        self._ready_future.result(10)
        super().start()

    def stop(self, *, force: bool = False) -> None:
        self._stopped = True
        if not force:
            self._thread.join(5)

        super().stop(force=force)

    def _listen_messages(self) -> None:
        while not self._stopped:
            try:
                pubsub = self.client.pubsub()
                pubsub.subscribe(self.channel)
            except BaseException as exc:
                if not self._ready_future.done():
                    self._ready_future.set_exception(exc)

                raise
            else:
                if not self._ready_future.done():
                    self._ready_future.set_result(None)

            try:
                while not self._stopped:
                    msg = pubsub.get_message(timeout=self.message_poll_interval)
                    if msg and isinstance(msg["data"], bytes):
                        event = self.reconstitute_event(msg["data"])
                        if event is not None:
                            self.publish_local(event)
            except BaseException:
                self._logger.exception("Subscriber crashed")
                raise
            finally:
                pubsub.close()

    def publish(self, event: Event) -> None:
        notification = self.generate_notification(event)
        self.client.publish(self.channel, notification)



"""
This module contains just the code for the scheduled task.
It should not be run directly.
"""

from __future__ import annotations

from datetime import datetime


def tick():
    print("Hello, the time is", datetime.now())



"""
Example demonstrating use of the asynchronous scheduler in a simple asyncio app.

To run: python async_memory.py

It should print a line on the console on a one-second interval.
"""

from __future__ import annotations

from asyncio import run
from datetime import datetime

from apscheduler.schedulers.async_ import AsyncScheduler
from apscheduler.triggers.interval import IntervalTrigger


def tick():
    print("Hello, the time is", datetime.now())


async def main():
    async with AsyncScheduler() as scheduler:
        await scheduler.add_schedule(tick, IntervalTrigger(seconds=1))
        await scheduler.wait_until_stopped()


run(main())



from __future__ import annotations

from abc import abstractmethod
from datetime import datetime, timedelta
from typing import Any

import attrs

from .._exceptions import MaxIterationsReached
from .._validators import as_timedelta, require_state_version
from ..abc import Trigger
from ..marshalling import marshal_object, unmarshal_object


@attrs.define
class BaseCombiningTrigger(Trigger):
    triggers: list[Trigger]
    _next_fire_times: list[datetime | None] = attrs.field(
        init=False, eq=False, factory=list
    )

    def __getstate__(self) -> dict[str, Any]:
        return {
            "version": 1,
            "triggers": [marshal_object(trigger) for trigger in self.triggers],
            "next_fire_times": self._next_fire_times,
        }

    @abstractmethod
    def __setstate__(self, state: dict[str, Any]) -> None:
        self.triggers = [
            unmarshal_object(*trigger_state) for trigger_state in state["triggers"]
        ]
        self._next_fire_times = state["next_fire_times"]


@attrs.define
class AndTrigger(BaseCombiningTrigger):
    """
    Fires on times produced by the enclosed triggers whenever the fire times are within
    the given threshold.

    If the produced fire times are not within the given threshold of each other, the
    trigger(s) that produced the earliest fire time will be asked for their next fire
    time and the iteration is restarted. If instead all the triggers agree on a fire
    time, all the triggers are asked for their next fire times and the earliest of the
    previously produced fire times will be returned.

    This trigger will be finished when any of the enclosed trigger has finished.

    :param triggers: triggers to combine
    :param threshold: maximum time difference between the next fire times of the
        triggers in order for the earliest of them to be returned from :meth:`next` (in
        seconds, or as timedelta)
    :param max_iterations: maximum number of iterations of fire time calculations before
        giving up
    """

    threshold: timedelta = attrs.field(converter=as_timedelta, default=1)
    max_iterations: int | None = 10000

    def next(self) -> datetime | None:
        if not self._next_fire_times:
            # Fill out the fire times on the first run
            self._next_fire_times = [t.next() for t in self.triggers]

        for _ in range(self.max_iterations):
            # Find the earliest and latest fire times
            earliest_fire_time: datetime | None = None
            latest_fire_time: datetime | None = None
            for fire_time in self._next_fire_times:
                # If any of the fire times is None, this trigger is finished
                if fire_time is None:
                    return None

                if earliest_fire_time is None or earliest_fire_time > fire_time:
                    earliest_fire_time = fire_time

                if latest_fire_time is None or latest_fire_time < fire_time:
                    latest_fire_time = fire_time

            # Replace all the fire times that were within the threshold
            for i, _trigger in enumerate(self.triggers):
                if self._next_fire_times[i] - earliest_fire_time <= self.threshold:
                    self._next_fire_times[i] = self.triggers[i].next()

            # If all the fire times were within the threshold, return the earliest one
            if latest_fire_time - earliest_fire_time <= self.threshold:
                self._next_fire_times = [t.next() for t in self.triggers]
                return earliest_fire_time
        else:
            raise MaxIterationsReached

    def __getstate__(self) -> dict[str, Any]:
        state = super().__getstate__()
        state["threshold"] = self.threshold.total_seconds()
        state["max_iterations"] = self.max_iterations
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        require_state_version(self, state, 1)
        super().__setstate__(state)
        self.threshold = timedelta(seconds=state["threshold"])
        self.max_iterations = state["max_iterations"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.triggers}, "
            f"threshold={self.threshold.total_seconds()}, "
            f"max_iterations={self.max_iterations})"
        )


@attrs.define
class OrTrigger(BaseCombiningTrigger):
    """
    Fires on every fire time of every trigger in chronological order.
    If two or more triggers produce the same fire time, it will only be used once.

    This trigger will be finished when none of the enclosed triggers can produce any new
    fire times.

    :param triggers: triggers to combine
    """

    def next(self) -> datetime | None:
        # Fill out the fire times on the first run
        if not self._next_fire_times:
            self._next_fire_times = [t.next() for t in self.triggers]

        # Find out the earliest of the fire times
        earliest_time: datetime | None = min(
            (fire_time for fire_time in self._next_fire_times if fire_time is not None),
            default=None,
        )
        if earliest_time is not None:
            # Generate new fire times for the trigger(s) that generated the earliest
            # fire time
            for i, fire_time in enumerate(self._next_fire_times):
                if fire_time == earliest_time:
                    self._next_fire_times[i] = self.triggers[i].next()

        return earliest_time

    def __setstate__(self, state: dict[str, Any]) -> None:
        require_state_version(self, state, 1)
        super().__setstate__(state)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.triggers})"
